/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Software Author: Tony Whyman mailto:tony.whyman@mwasoftware.co.uk
   http://www.mwasoftware.co.uk


   Change Log:


   14/5/01 Added warning check on IBServer running - see CheckIBServer

   16/9/02 Major Revisions including removal of CheckIBServer, IsIBRunning
   		  now sets a property instead of returing an Error result. New
           function "IBStop" to stop a running server.

*/

#include <stdlib.h>
#include <windows.h>
#include <msiquery.h>  //Supplied as part of the MS Platform SDK
#include <ver.h>
#include <dir.h>
#include <stdio.h>

//Error numbers used as keys into the Merge Module's Error Table

#define NOSERVICEDBPATH 27014
#define NOVERSIONINFORMATION 27015
#define GETPROPERTYERROR 27016
#define NOTSYSTEMDIR 27017
#define INTERBASE65 27019
#define SETPROPERTYERROR 27020
#define REGKEYERROR 27021
#define FILENOTPRESENT 27022
#define OLDVERSION NOVERSIONINFORMATION
#define OUTOFMEMORY 17

//Property Names
#define IBRUNNING "IBRUNNING.63C0C361_3142_4765_98CC_A503E95477CD"
#define IB5X "IB5X.2293ED4E_0525_49AA_84BF_6F52B52BB20D"
#define IB6 "IB6.2293ED4E_0525_49AA_84BF_6F52B52BB20D"
#define IBINSTALLPATH "IBINSTALLPATH.2293ED4E_0525_49AA_84BF_6F52B52BB20D"
#define IBLATERVERSION "IBLATERVERSION.2293ED4E_0525_49AA_84BF_6F52B52BB20D"

//Registry Keys
#define REGKEYSHAREDLL "Software\\Microsoft\\Windows\\CurrentVersion\\SharedDLLs"

//The following is copied from iscguard/iscguard.rh in the InterBase source
#define IDM_SHUTDOWN 5


//-----------------------------------------------------------------------------
//Display a error message with no parameters

void Error(MSIHANDLE hInstall, int Message){
MSIHANDLE hRec;

  hRec = MsiCreateRecord(2);
  __try {
    MsiRecordSetInteger(hRec,1,Message);
    MsiProcessMessage(hInstall,INSTALLMESSAGE_ERROR,hRec);
  }
  __finally {
    MsiCloseHandle(hRec);
  }
}

//-----------------------------------------------------------------------------
//Display a Fatal Error Message with no parameters

void FatalError(MSIHANDLE hInstall, int Message){
MSIHANDLE hRec;

  hRec = MsiCreateRecord(2);
  __try {
    MsiRecordSetInteger(hRec,1,Message);
    MsiProcessMessage(hInstall,INSTALLMESSAGE_FATALEXIT,hRec);
  }
  __finally {
    MsiCloseHandle(hRec);
  }
}

//-----------------------------------------------------------------------------
//Display an error message with the "Filename" as the parameter

void Error1(MSIHANDLE hInstall, int Message, LPTSTR Filename){
MSIHANDLE hRec;

  hRec = MsiCreateRecord(3);
  __try {
    MsiRecordSetInteger(hRec,1,Message);
    MsiRecordSetString(hRec,2,(LPCTSTR) Filename);
    MsiProcessMessage(hInstall,INSTALLMESSAGE_ERROR,hRec);
  }
  __finally {
    MsiCloseHandle(hRec);
  }
}

//-----------------------------------------------------------------------------
//Display a Warning Message with no parameters

void Warning(MSIHANDLE hInstall, int Message){
MSIHANDLE hRec;

  hRec = MsiCreateRecord(2);
  __try {
    MsiRecordSetInteger(hRec,1,Message);
    MsiProcessMessage(hInstall,INSTALLMESSAGE_WARNING,hRec);
  }
  __finally {
    MsiCloseHandle(hRec);
  }
}

//-----------------------------------------------------------------------------
//If the IB Guardian or Server is running then display an error message and
//force termination of the installation

UINT __stdcall IsIBRunning(MSIHANDLE hInstall)
{
  if (FindWindow("IB_Server", "InterBase Server")
    || FindWindow("IB_Guard", "InterBase Guardian")) {

   MsiSetProperty(hInstall,IBRUNNING,"Yes");
  }
  return ERROR_SUCCESS;

}
//-----------------------------------------------------------------------------
//Locate and stop the InterBase or Firebird Server

UINT __stdcall IBStop(MSIHANDLE hInstall)
{
  HWND hTmpWnd;
  hTmpWnd = FindWindow("IB_Guard", "InterBase Guardian");
  if (hTmpWnd == NULL)
    hTmpWnd = FindWindow("IB_Server", "InterBase Server");
  if (hTmpWnd != NULL)
    PostMessage(hTmpWnd, WM_COMMAND, (WPARAM) IDM_SHUTDOWN, 0);
  return ERROR_SUCCESS;

}
//-----------------------------------------------------------------------------
//Parse the path string returning the next path in the list, while setting
//the end of string marker for the first path in the list.

char *GetNextPath(char *Path){

  char *P = strchr(Path,';');
  if (P) {
    *P = '\0';
    return P+1;
  }
  return NULL;
}
//-----------------------------------------------------------------------------
// Returns 1 if FileName is a valid path to an existing file and 0 otherwise

int FileExists(char *FileName) {

WIN32_FIND_DATA Data;
HANDLE SearchHandle;

  if ((SearchHandle = FindFirstFile(FileName,&Data))
    == INVALID_HANDLE_VALUE)
    return 0;

  FindClose(SearchHandle);
  return 1;
}
//-----------------------------------------------------------------------------
//Returns the Directory path of FileName in the Directory buffer - which is
//assumed large enough to contain the directory path.

LPTSTR GetDirectory(LPTSTR FileName, LPTSTR Directory) {
char drive[MAXDRIVE];
char dir[MAXDIR];

  fnsplit(FileName,drive,dir,NULL,NULL);
  while (strlen(dir) && (dir[strlen(dir)-1] == '\\')) {
    dir[strlen(dir)-1] = '\0';
  }
  strcat(strcpy(Directory,drive),dir);
  return Directory;
}

//-----------------------------------------------------------------------------
//Checks the Version information in "Filename" which is assumed to be a path
//to a copy of gds32.dll. Displays an error message if no version information
//older than version 5 or not in the SystemFolder.

UINT CheckGDS32(MSIHANDLE hInstall, LPTSTR Filename)
{
  DWORD Size;
  DWORD Handle;
  void* mem;
  void* FixedFileInfo;
  UINT InfoBytes;
  char Directory[MAXPATH];
  char SysDir[MAXPATH];

  if ((Size = GetFileVersionInfoSize(Filename,&Handle)) == 0) {
    Error1(hInstall,NOVERSIONINFORMATION,Filename);
    return ERROR_INSTALL_FAILURE;
  }

  if ((mem = malloc(Size)) == NULL) {
    FatalError(hInstall,OUTOFMEMORY);
    return ERROR_INSTALL_FAILURE;
  }

  __try {
    if (!GetFileVersionInfo(Filename,Handle,Size,mem)) {
      Error1(hInstall,NOVERSIONINFORMATION,Filename);
      return ERROR_INSTALL_FAILURE;
    }

    if (!VerQueryValue(mem,TEXT("\\"),&FixedFileInfo,&InfoBytes)) {
      Error1(hInstall,NOVERSIONINFORMATION,Filename);
      return ERROR_INSTALL_FAILURE;
    }

    if (HIWORD(((VS_FIXEDFILEINFO*)FixedFileInfo)->dwFileVersionMS) <= 4) {
      Error1(hInstall,OLDVERSION,Filename);
      return ERROR_INSTALL_FAILURE;
    }

    if (HIWORD(((VS_FIXEDFILEINFO*)FixedFileInfo)->dwFileVersionMS) == 5)
      MsiSetProperty(hInstall,IB5X,"Yes");
    else
    if (HIWORD(((VS_FIXEDFILEINFO*)FixedFileInfo)->dwFileVersionMS) == 6 &&
        LOWORD(((VS_FIXEDFILEINFO*)FixedFileInfo)->dwFileVersionMS) == 0)
      MsiSetProperty(hInstall,IB6,"Yes");

    if ((HIWORD(((VS_FIXEDFILEINFO*)FixedFileInfo)->dwFileVersionMS) == 6 &&
        LOWORD(((VS_FIXEDFILEINFO*)FixedFileInfo)->dwFileVersionMS) == 5)
        || HIWORD(((VS_FIXEDFILEINFO*)FixedFileInfo)->dwFileVersionMS) > 6) {
      char buffer[64];
      wsprintf(buffer,"%d.%d.%d.%d",
               HIWORD(((VS_FIXEDFILEINFO*)FixedFileInfo)->dwFileVersionMS),
               LOWORD(((VS_FIXEDFILEINFO*)FixedFileInfo)->dwFileVersionMS),
               HIWORD(((VS_FIXEDFILEINFO*)FixedFileInfo)->dwFileVersionLS),
               LOWORD(((VS_FIXEDFILEINFO*)FixedFileInfo)->dwFileVersionLS));
      MsiSetProperty(hInstall,IBLATERVERSION,buffer);
      return ERROR_SUCCESS;
    }
  }
  __finally {
    free(mem);
  }

  GetSystemDirectory(&SysDir,MAXPATH);
  GetDirectory(Filename,&Directory);
  if (strcmpi(&Directory,&SysDir) != 0) {
    Error1(hInstall,NOTSYSTEMDIR,&Directory);
    return ERROR_INSTALL_FAILURE;
  }
  return ERROR_SUCCESS;
}

//-----------------------------------------------------------------------------
// Calls CheckGDS32 for the filename given by the Installation Script Property
//  "GDS32_FN1", and returns the response.

UINT __stdcall CheckGDS32_1(MSIHANDLE hInstall) {

char Filename[MAXPATH];
DWORD count = MAXPATH;

  if (MsiGetProperty(hInstall,"GDS32_FN1",&Filename,&count) != ERROR_SUCCESS) {
    Error1(hInstall,GETPROPERTYERROR,"GDS32_FN1");
    return ERROR_INSTALL_FAILURE;
  }

  return CheckGDS32(hInstall,Filename);
}

//-----------------------------------------------------------------------------
// Calls CheckGDS32 for the filename given by the Installation Script Property
//  "GDS32_FN2", and returns the response.

UINT __stdcall CheckGDS32_2(MSIHANDLE hInstall) {

char Filename[MAXPATH];
DWORD count = MAXPATH;

  if (MsiGetProperty(hInstall,"GDS32_FN2",&Filename,&count) != ERROR_SUCCESS) {
    Error1(hInstall,GETPROPERTYERROR,"GDS32_FN2");
    return ERROR_INSTALL_FAILURE;
  }

  return CheckGDS32(hInstall,Filename);
}

//-----------------------------------------------------------------------------
// Calls CheckGDS32 for the filename given by the Installation Script Property
//  "GDS32_FN3", and returns the response.

UINT __stdcall CheckGDS32_3(MSIHANDLE hInstall) {

char Filename[MAXPATH];
DWORD count = MAXPATH;

  if (MsiGetProperty(hInstall,"GDS32_FN3",&Filename,&count) != ERROR_SUCCESS) {
    Error1(hInstall,GETPROPERTYERROR,"GDS32_FN3");
    return ERROR_INSTALL_FAILURE;
  }

  return CheckGDS32(hInstall,Filename);
}


//-----------------------------------------------------------------------------
//Searches the "PathList" for gds32.dll. For each copy found, CheckGDS32 is used
//to determine the if there is a problem installing over the copy. If CheckGDS32
//returns an error response then the search stops and the response is returned.

UINT SearchPathList(MSIHANDLE hInstall, char *PathList) {

char *Next = NULL;
char *Path = PathList;
char FileName[MAXPATH];
UINT response;

 if (PathList)
  do {
    Next = GetNextPath(Path);
    strcat(strcat(strcpy(FileName,Path),"\\"),"gds32.dll");
    if (FileExists(FileName)) {
      if ((response = CheckGDS32(hInstall,FileName)) != ERROR_SUCCESS)
        return response;
    }
    Path = Next;
  } while (Path != NULL);
 return ERROR_SUCCESS;
}
//-----------------------------------------------------------------------------
// Calls SearchPathList with the value of the Path environment variable
// as the SeachPath, and returns the response.

UINT __stdcall Search4GDS32(MSIHANDLE hInstall) {

  return SearchPathList(hInstall,getenv("Path"));

}

//-----------------------------------------------------------------------------
// Work out where the Services Database is.

int GetServiceDBPath(char *ServiceDBPath) {

char WinDir[MAXPATH];

  GetWindowsDirectory(&WinDir,MAXPATH);

  //Try Win9x first

  if (FileExists(strcat(strcpy(ServiceDBPath,&WinDir),"\\Services")))
    return 1;
  // try WinNT

  if (FileExists(strcat(strcpy(ServiceDBPath,&WinDir),"\\System32\\drivers\\etc\\Services")))
    return 1;

  return 0;

}

//-----------------------------------------------------------------------------
//Check the Services file for th gds_db entry and add one if not found.

void AddGDSEntry(char *ServiceDBPath) {

FILE *Services;
char line[128];

  if ((Services = fopen(ServiceDBPath,"rt")) != NULL) {
    __try {
      while (fgets(&line,128,Services)) {
        if (strncmp(&line,"gds_db",6) == 0) return;
      }
    }
    __finally {
      fclose(Services);
    }
  }

  if ((Services = fopen(ServiceDBPath,"at")) != NULL) {
    __try {
      fputs("\ngds_db\t3050/tcp\n",Services);
    }
    __finally {
      fclose(Services);
    }
  }
}

//-----------------------------------------------------------------------------
//Determines where the Services file is and then calls AddGDSEntry.

UINT __stdcall AddService(MSIHANDLE hInstall) {

char ServiceDBPath[MAXPATH];

  if (GetServiceDBPath(&ServiceDBPath)) {
    AddGDSEntry(&ServiceDBPath);
  }
  else
   Error(hInstall,NOSERVICEDBPATH);
  return ERROR_SUCCESS;
}


//-----------------------------------------------------------------------------
//Set CurrentVersion UseCount or Remove if down to zero
//
void SetCurValueUseCount(HKEY Key, int up){

DWORD UseCount;

  if (RegOpenKeyEx(Key,"CurrentVersion",0,KEY_READ | KEY_WRITE,&Key) == ERROR_SUCCESS) {
    DWORD Count = sizeof(UseCount);
    if (RegQueryValueEx(Key,"UseCount",NULL,NULL,&UseCount,&Count) == ERROR_SUCCESS) {
      if (up) UseCount++; else UseCount--;
      if (UseCount == 0) {
        RegDeleteValue(Key,"UseCount");
        return;
      }
    }
    else if (up) UseCount = 1;
    else return;
    RegSetValueEx(Key,"UseCount",NULL,REG_DWORD,&UseCount,sizeof(UseCount));
  }
}

//-----------------------------------------------------------------------------
//Set InterBase UseCount or Remove if down to zero
//
void SetInterBaseUseCount(HKEY Key, int up){

DWORD UseCount;

  if (RegOpenKeyEx(Key,"InterBase",0,KEY_READ | KEY_WRITE,&Key) == ERROR_SUCCESS) {
    DWORD Count = sizeof(UseCount);
    SetCurValueUseCount(Key,up);
    if (RegQueryValueEx(Key,"UseCount",NULL,NULL,&UseCount,&Count) == ERROR_SUCCESS) {
      if (up) UseCount++; else UseCount--;
      if (UseCount == 0) {
        RegDeleteValue(Key,"UseCount");
        return;
      }
    }
    else if (up) UseCount = 1;
    else return;
    RegSetValueEx(Key,"UseCount",NULL,REG_DWORD,&UseCount,sizeof(UseCount));
  }
}

//-----------------------------------------------------------------------------
//Set Company Use Count or Remove if down to zero
//
void SetCompanyUseCounts(char *CompanyName, int up) {

HKEY Key;
DWORD UseCount = 1;
char buf[128];

  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,strcat(strcpy(&buf,"SOFTWARE\\"),CompanyName),0,KEY_READ | KEY_WRITE,&Key) == ERROR_SUCCESS) {
    DWORD Count = sizeof(UseCount);
    SetInterBaseUseCount(Key,up);
    if (RegQueryValueEx(Key,"UseCount",NULL,NULL,&UseCount,&Count) == ERROR_SUCCESS) {
      if (up) UseCount++; else UseCount--;
      if (UseCount == 0) {
        RegDeleteValue(Key,"UseCount");
        return;
      }
    }
    else if (up) UseCount = 1;
    else return;
    RegSetValueEx(Key,"UseCount",NULL,REG_DWORD,&UseCount,sizeof(UseCount));
  }
}

//-----------------------------------------------------------------------------
//Increment Use Counts
//
UINT __stdcall IncUseCounts(MSIHANDLE hInstall) {
  SetCompanyUseCounts("InterBase Corp",1);
  SetCompanyUseCounts("Borland",1);
  return ERROR_SUCCESS;
}

//-----------------------------------------------------------------------------
//Decrement Use Counts
UINT __stdcall DecUseCounts(MSIHANDLE hInstall) {
  SetCompanyUseCounts("InterBase Corp",0);
  SetCompanyUseCounts("Borland",0);
  return ERROR_SUCCESS;
}
//-----------------------------------------------------------------------------
//Register use of existing shared DLL

UINT RegisterDLL(MSIHANDLE hInstall, char *DLLName){

  HKEY Key;
  DWORD UseCount;
  DWORD Count = sizeof(UseCount);
  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,REGKEYSHAREDLL,0,KEY_READ | KEY_WRITE,&Key) == ERROR_SUCCESS) {
    if (RegQueryValueEx(Key,DLLName,NULL,NULL,&UseCount,&Count) == ERROR_SUCCESS) {
      UseCount++;
      RegSetValueEx(Key,DLLName,NULL,REG_DWORD,&UseCount,sizeof(UseCount));
    }
    else {
      Error1(hInstall,FILENOTPRESENT,DLLName);
      return ERROR_INSTALL_FAILURE;
    }
  }
  else {
    FatalError(hInstall,REGKEYERROR);
    return ERROR_INSTALL_FAILURE;
  }
  return ERROR_SUCCESS;
}
//-----------------------------------------------------------------------------
//Deregister use of shared DLL

UINT DeRegisterDLL(MSIHANDLE hInstall, char *DLLName){

  HKEY Key;
  DWORD UseCount;
  DWORD Count = sizeof(UseCount);
  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE,REGKEYSHAREDLL,0,KEY_READ | KEY_WRITE,&Key) == ERROR_SUCCESS) {
    if (RegQueryValueEx(Key,DLLName,NULL,NULL,&UseCount,&Count) == ERROR_SUCCESS) {
      UseCount--;
      if (UseCount == 0) {
        RegDeleteValue(Key,DLLName);
        DeleteFile(DLLName);
      }
      else
        RegSetValueEx(Key,DLLName,NULL,REG_DWORD,&UseCount,sizeof(UseCount));
    }
  }
  else {
    FatalError(hInstall,REGKEYERROR);
    return ERROR_INSTALL_FAILURE;
  }
  return ERROR_SUCCESS;
}
//-----------------------------------------------------------------------------
//Register Client Files (already installed by IB6)

UINT __stdcall RegisterClientFiles(MSIHANDLE hInstall) {

  char InstallDir[MAXPATH];
  char DLLPath[MAXPATH];
  char SysDir[MAXPATH];
  UINT ReturnCode;

  GetSystemDirectory(&SysDir,MAXPATH);
  ReturnCode = RegisterDLL(hInstall,strcat(strcpy(DLLPath,SysDir),"\\gds32.dll"));
  if (ReturnCode == ERROR_SUCCESS) {
    DWORD count = sizeof(InstallDir);
    if (MsiGetProperty(hInstall,IBINSTALLPATH,&InstallDir,&count) != ERROR_SUCCESS) {
      Error1(hInstall,GETPROPERTYERROR,IBINSTALLPATH);
      return ERROR_INSTALL_FAILURE;
    }

    return RegisterDLL(hInstall,strcat(strcpy(DLLPath,InstallDir),"interbase.msg"));
  }
  return ReturnCode;
}
//-----------------------------------------------------------------------------
//DeRegister Client Files (already installed by IB6)

UINT __stdcall DeRegisterClientFiles(MSIHANDLE hInstall) {

  char InstallDir[MAXPATH];
  char DLLPath[MAXPATH];
  char SysDir[MAXPATH];
  UINT ReturnCode;
  DWORD count = sizeof(InstallDir);

  if (MsiGetProperty(hInstall,IBINSTALLPATH,&InstallDir,&count) != ERROR_SUCCESS) {
    Error1(hInstall,GETPROPERTYERROR,IBINSTALLPATH);
    return ERROR_INSTALL_FAILURE;
  }

  ReturnCode = DeRegisterDLL(hInstall,strcat(strcpy(DLLPath,InstallDir),"interbase.msg"));
  if (ReturnCode == ERROR_SUCCESS) {
    GetSystemDirectory(&SysDir,MAXPATH);
    return DeRegisterDLL(hInstall,strcat(strcpy(DLLPath,SysDir),"\\gds32.dll"));
  }
  return ReturnCode;
}
//-----------------------------------------------------------------------------
//Register Server Files (already installed by IB6)

UINT __stdcall RegisterServerFiles(MSIHANDLE hInstall) {

  char InstallDir[MAXPATH];
  char DLLPath[MAXPATH];
  UINT ReturnCode;
  DWORD count = sizeof(InstallDir);

  if (MsiGetProperty(hInstall,IBINSTALLPATH,&InstallDir,&count) != ERROR_SUCCESS) {
    Error1(hInstall,GETPROPERTYERROR,IBINSTALLPATH);
    return ERROR_INSTALL_FAILURE;
  }

  ReturnCode =  RegisterDLL(hInstall,strcat(strcpy(DLLPath,InstallDir),"bin\\ibguard.exe"));
  if (ReturnCode == ERROR_SUCCESS)
    ReturnCode = RegisterDLL(hInstall,strcat(strcpy(DLLPath,InstallDir),"bin\\ibserver.exe"));
  if (ReturnCode == ERROR_SUCCESS)
    ReturnCode = RegisterDLL(hInstall,strcat(strcpy(DLLPath,InstallDir),"intl\\gdsintl.dll"));
  return ReturnCode;
}
//-----------------------------------------------------------------------------
//DeRegister Server Files (already installed by IB6)

UINT __stdcall DeRegisterServerFiles(MSIHANDLE hInstall) {
  char InstallDir[MAXPATH];
  char DLLPath[MAXPATH];
  UINT ReturnCode;
  DWORD count = sizeof(InstallDir);

  if (MsiGetProperty(hInstall,IBINSTALLPATH,&InstallDir,&count) != ERROR_SUCCESS) {
    Error1(hInstall,GETPROPERTYERROR,IBINSTALLPATH);
    return ERROR_INSTALL_FAILURE;
  }

  ReturnCode = DeRegisterDLL(hInstall,strcat(strcpy(DLLPath,InstallDir),"bin\\ibguard.exe"));
  if (ReturnCode == ERROR_SUCCESS)
    ReturnCode = DeRegisterDLL(hInstall,strcat(strcpy(DLLPath,InstallDir),"bin\\ibserver.exe"));
  if (ReturnCode == ERROR_SUCCESS)
    ReturnCode = DeRegisterDLL(hInstall,strcat(strcpy(DLLPath,InstallDir),"intl\\gdsintl.dll"));
  return ReturnCode;
}

